;; Generated package description from colorful-mode.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "colorful-mode" "1.2.5.0.20260424.194108" "Preview any color in your buffer in real time" '((emacs "28.1") (compat "30.1.0.0")) :commit "13622ae3deeb911aecf827f51aa30679de15e773" :authors '(("Elias G. Perez" . "eg642616@gmail.com")) :maintainer '(("Jen-Chieh" . "jcs090218@gmail.com") ("Elias G. Perez" . "eg642616@gmail.com")) :keywords '("faces" "tools" "matching" "convenience") :url "https://github.com/DevelopmentCool2449/colorful-mode")
